package eu.jonahbauer.raytracing.render.color;

import org.jetbrains.annotations.NotNull;

public final class TransferFunctions {
    public static final @NotNull TransferFunction sRGB = new ComponentTransferFunction() {
        @Override
        protected double encode(double value) {
            if (value <= 0.0031308) return 12.92 * value;
            return 1.055 * Math.pow(value, 1. / 2.4) - 0.055;
        }

        @Override
        protected double decode(double value) {
            if (value <= 0.04045) return value / 12.92;
            return Math.pow((value + 0.055) / 1.055, 2.4d);
        }
    };

    public static final @NotNull TransferFunction LINEAR = new TransferFunction() {
        @Override
        public @NotNull ColorRGB encode(@NotNull ColorRGB rgb) {
            return rgb;
        }

        @Override
        public @NotNull ColorRGB decode(@NotNull ColorRGB rgb) {
            return rgb;
        }
    };

    private TransferFunctions() {
        throw new UnsupportedOperationException();
    }

    private abstract static class ComponentTransferFunction implements TransferFunction {
        @Override
        public final @NotNull ColorRGB decode(@NotNull ColorRGB rgb) {
            return new ColorRGB(decode(rgb.r()), decode(rgb.g()), decode(rgb.b()));
        }

        @Override
        public final @NotNull ColorRGB encode(@NotNull ColorRGB rgb) {
            return new ColorRGB(encode(rgb.r()), encode(rgb.g()), encode(rgb.b()));
        }

        protected abstract double encode(double value);
        protected abstract double decode(double value);
    }
}
